package gov.nasa.ziggy.supervisor;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;
import static org.mockito.Mockito.times;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;
import org.mockito.ArgumentMatchers;
import org.mockito.Mockito;

import gov.nasa.ziggy.ZiggyDatabaseRule;
import gov.nasa.ziggy.pipeline.PipelineExecutor;
import gov.nasa.ziggy.pipeline.definition.PipelineInstance;
import gov.nasa.ziggy.pipeline.definition.PipelineInstanceNode;
import gov.nasa.ziggy.pipeline.definition.PipelineTask;
import gov.nasa.ziggy.pipeline.definition.database.PipelineInstanceNodeOperations;
import gov.nasa.ziggy.pipeline.definition.database.PipelineOperationsTestUtils;
import gov.nasa.ziggy.pipeline.definition.database.PipelineTaskDataOperations;
import gov.nasa.ziggy.pipeline.definition.database.PipelineTaskOperations;
import gov.nasa.ziggy.pipeline.step.remote.BatchManager;
import gov.nasa.ziggy.services.alert.Alert.Severity;
import gov.nasa.ziggy.services.alert.AlertService;
import gov.nasa.ziggy.services.messages.HaltTasksRequest;
import gov.nasa.ziggy.services.messages.RetryTransitionRequest;
import gov.nasa.ziggy.services.messages.TaskHaltedMessage;
import gov.nasa.ziggy.util.PipelineException;

/**
 * Unit tests for {@link PipelineSupervisor} class.
 * <p>
 * Given the literal centrality of the supervisor to everything Ziggy does, you would expect that we
 * would test it to within an inch of its life. However: Most of the supervisor's functions are
 * derived from the worker class in Spiffy, in which context there were exactly zero unit tests.
 * Hence, for the time being, all we can really manage is to add tests for new functionality and
 * hope that someday we get the chance to backfill tests for legacy functionality.
 *
 * @author PT
 * @author Bill Wohler
 */
public class PipelineSupervisorTest {

    @Rule
    public ZiggyDatabaseRule databaseRule = new ZiggyDatabaseRule();

    private PipelineSupervisor supervisor;
    private PipelineTaskOperations pipelineTaskOperations = Mockito
        .mock(PipelineTaskOperations.class);
    private PipelineTaskDataOperations pipelineTaskDataOperations = Mockito
        .mock(PipelineTaskDataOperations.class);
    private AlertService alertService = Mockito.mock(AlertService.class);
    private PipelineExecutor pipelineExecutor = Mockito.mock(PipelineExecutor.class);
    private Map<PipelineTask, List<Long>> jobIdsByTask = new HashMap<>();
    private List<PipelineTask> tasks = new ArrayList<>();
    private BatchManager<?> batchManager = Mockito.mock(BatchManager.class);

    private PipelineTask pipelineTask1;
    private PipelineTask pipelineTask2;
    private PipelineTask pipelineTask3;
    private PipelineTask pipelineTask4;

    @Before
    public void setUp() {
        supervisor = Mockito.spy(new PipelineSupervisor(1, 12000));
        Mockito.doReturn(alertService).when(supervisor).alertService();
        Mockito.doReturn(pipelineTaskOperations).when(supervisor).pipelineTaskOperations();
        Mockito.doReturn(pipelineTaskDataOperations).when(supervisor).pipelineTaskDataOperations();
        Mockito.doReturn(jobIdsByTask).when(supervisor).jobIdsByTask(ArgumentMatchers.anyList());
        Mockito.doReturn(pipelineExecutor).when(supervisor).pipelineExecutor();
        Mockito.doReturn(batchManager)
            .when(supervisor)
            .batchManager(ArgumentMatchers.any(PipelineTask.class));

        pipelineTask1 = Mockito.spy(PipelineTask.class);
        Mockito.doReturn(1L).when(pipelineTask1).getId();
        pipelineTask2 = Mockito.spy(PipelineTask.class);
        Mockito.doReturn(2L).when(pipelineTask2).getId();
        pipelineTask3 = Mockito.spy(PipelineTask.class);
        Mockito.doReturn(3L).when(pipelineTask3).getId();
        pipelineTask4 = Mockito.spy(PipelineTask.class);
        Mockito.doReturn(4L).when(pipelineTask4).getId();
    }

    /** Verify that the task-halt command works when no remote tasks are present. */
    @Test
    public void testHaltTasksEmptyStateFileCollection() {
        supervisor.haltRemoteTasks(
            new HaltTasksRequest(Set.of(pipelineTask1, pipelineTask2, pipelineTask3)));
        Mockito.verify(supervisor, times(0))
            .publishTaskHaltedMessage(ArgumentMatchers.any(PipelineTask.class));
    }

    /**
     * Verify that the halt-tasks command works when the state file collection and the halt request
     * have some overlap in tasks but not complete (specifically, each of them has a task that the
     * other one does not).
     */
    @Test
    public void testHaltTasksAllSuccessful() {

        jobIdsByTask.put(pipelineTask1, List.of(100L, 101L));
        jobIdsByTask.put(pipelineTask2, List.of(102L, 103L));
        jobIdsByTask.put(pipelineTask3, List.of(105L, 106L));

        tasks.add(pipelineTask1);
        tasks.add(pipelineTask2);
        tasks.add(pipelineTask4);

        Mockito.when(batchManager.deleteJobs(pipelineTask1)).thenReturn(0);
        Mockito.when(batchManager.deleteJobs(pipelineTask2)).thenReturn(0);

        HaltTasksRequest request = new HaltTasksRequest(tasks);
        supervisor.haltRemoteTasks(request);
        Mockito.verify(supervisor).publishTaskHaltedMessage(pipelineTask1);
        Mockito.verify(supervisor).publishTaskHaltedMessage(pipelineTask2);
        Mockito.verify(supervisor, times(0)).publishTaskHaltedMessage(pipelineTask3);
        Mockito.verify(supervisor, times(0)).publishTaskHaltedMessage(pipelineTask4);
    }

    /**
     * Verify that the correct action is taken when not all the tasks to be halted are in fact
     * successfully halted.
     */
    @Test
    public void testHaltTasksSomeFailures() {
        jobIdsByTask.put(pipelineTask1, List.of(100L, 101L));
        jobIdsByTask.put(pipelineTask2, List.of(102L, 203L));
        jobIdsByTask.put(pipelineTask3, List.of(105L, 106L));

        tasks.add(pipelineTask1);
        tasks.add(pipelineTask2);
        tasks.add(pipelineTask4);

        Mockito.when(batchManager.deleteJobs(pipelineTask1)).thenReturn(0);
        Mockito.when(batchManager.deleteJobs(pipelineTask2)).thenReturn(1);

        // Two kill task requests should have gone out.
        HaltTasksRequest request = new HaltTasksRequest(tasks);
        supervisor.haltRemoteTasks(request);

        // Only the case where qdel returned 0 should publish a task message.
        Mockito.verify(supervisor).publishTaskHaltedMessage(pipelineTask1);
        Mockito.verify(supervisor, times(0)).publishTaskHaltedMessage(pipelineTask2);
        Mockito.verify(supervisor, times(0)).publishTaskHaltedMessage(pipelineTask3);
        Mockito.verify(supervisor, times(0)).publishTaskHaltedMessage(pipelineTask4);
    }

    /** Verify that the correct action is taken when a {@link TaskHaltedMessage} arrives. */
    @Test
    public void testHandleHaltedTaskMessage() {
        assertFalse(PipelineSupervisor.taskOnHaltedTaskList(pipelineTask1));
        TaskHaltedMessage message = new TaskHaltedMessage(pipelineTask1);
        supervisor.handleTaskHaltedMessage(message);
        assertTrue(PipelineSupervisor.taskOnHaltedTaskList(pipelineTask1));
        Mockito.verify(alertService)
            .generateAndBroadcastAlert("PI", pipelineTask1, Severity.ERROR, "Task 1 halted");
        Mockito.verify(pipelineTaskDataOperations).taskErrored(pipelineTask1);
    }

    @Test
    public void testRetryTransition() {
        PipelineOperationsTestUtils testUtils = new PipelineOperationsTestUtils();
        testUtils.setUpSingleNodePipeline();
        PipelineInstanceNode pipelineInstanceNode = testUtils.pipelineInstanceNode();
        PipelineInstance pipelineInstance = testUtils.pipelineInstance();

        // set the failed-transition state.
        new PipelineInstanceNodeOperations().markInstanceNodeTransitionFailed(pipelineInstanceNode);

        RetryTransitionRequest retryTransitionRequest = new RetryTransitionRequest(
            pipelineInstance.getId());
        supervisor.retryTransition(retryTransitionRequest);

        PipelineInstanceNode nodeFromDatabase = new PipelineInstanceNodeOperations()
            .pipelineInstanceNode(pipelineInstanceNode.getId());
        assertFalse(nodeFromDatabase.isTransitionFailed());
        assertFalse(nodeFromDatabase.isTransitionComplete());
        assertEquals(PipelineInstance.State.PROCESSING,
            new PipelineInstanceNodeOperations().pipelineInstance(pipelineInstance.getId())
                .getState());

        Mockito.verify(pipelineExecutor).transitionToNextInstanceNode(pipelineInstanceNode);
    }

    @Test(expected = PipelineException.class)
    public void testRetryTransitionNoFailedInstances() {
        PipelineOperationsTestUtils testUtils = new PipelineOperationsTestUtils();
        testUtils.setUpSingleNodePipeline();
        PipelineInstance pipelineInstance = testUtils.pipelineInstance();
        RetryTransitionRequest retryTransitionRequest = new RetryTransitionRequest(
            pipelineInstance.getId());
        supervisor.retryTransition(retryTransitionRequest);
    }

    @Test(expected = PipelineException.class)
    public void testRetryTransitionTooManyFailedInstances() {
        PipelineOperationsTestUtils testUtils = new PipelineOperationsTestUtils();
        testUtils.setUpFourNodePipelineWithInstanceNodes();
        List<PipelineInstanceNode> pipelineInstanceNodes = testUtils.getPipelineInstanceNodes();
        PipelineInstance pipelineInstance = testUtils.pipelineInstance();
        PipelineInstanceNodeOperations pipelineInstanceNodeOperations = new PipelineInstanceNodeOperations();
        pipelineInstanceNodeOperations
            .markInstanceNodeTransitionFailed(pipelineInstanceNodes.get(0));
        pipelineInstanceNodeOperations
            .markInstanceNodeTransitionFailed(pipelineInstanceNodes.get(1));
        RetryTransitionRequest retryTransitionRequest = new RetryTransitionRequest(
            pipelineInstance.getId());
        supervisor.retryTransition(retryTransitionRequest);
    }
}
